"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const metadata_1 = require("./metadata/metadata");
const client_1 = require("./client");
const fs = require("fs");
const path = require("path");
const logger_1 = require("./logger/logger");
const logger_client_1 = require("./client/logger.client");
class Redisk {
    constructor(metadata, client) {
        this.metadata = metadata;
        this.client = client;
    }
    static init(options) {
        logger_1.default.info('Starting Redisk with default constructor...');
        return new Redisk(new metadata_1.Metadata(), new logger_client_1.LoggerClient(new client_1.RedisClient(options)));
    }
    close() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.client.closeConnection();
        });
    }
    getClient() {
        return this.client;
    }
    save(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, uniques, primary, canBeListed, properties, hasOneRelations } = this.metadata.getEntityMetadataFromInstance(entity);
            const hashKey = name + ':' + entity[primary];
            const persistedEntity = yield this.getOne(entity.constructor, entity[primary]);
            if (persistedEntity !== null) {
                const changedFields = [];
                for (const property of Object.keys(properties).map(key => properties[key])) {
                    if (entity[property.name] === undefined) {
                        entity[property.name] = persistedEntity[property.name];
                    }
                    if (entity[property.name] !== persistedEntity[property.name]) {
                        changedFields.push(property.name);
                        if (entity[property.name] === null) {
                            yield this.client.hdel(hashKey, property.name);
                        }
                        if (hasOneRelations !== undefined && hasOneRelations[property.name] && hasOneRelations[property.name].cascadeUpdate && entity[property.name] !== null) {
                            yield this.save(entity[property.name]);
                        }
                        if (property.searchable && persistedEntity[property.name]) {
                            yield this.client.srem(this.getSearchableKeyName(name, property.name), this.getSearchableValuePrefix(entity[primary]) + persistedEntity[property.name].toLowerCase());
                        }
                        if (property.indexed) {
                            yield this.dropIndex(persistedEntity, property, persistedEntity[primary]);
                        }
                    }
                }
                if (uniques) {
                    const uniquesChanged = changedFields.some(value => uniques.indexOf(value) >= 0);
                    if (uniquesChanged) {
                        yield this.dropUniqueKeys(persistedEntity);
                    }
                }
            }
            if (uniques) {
                for (const uniqueName of uniques) {
                    const entityWithUnique = yield this.getOne(entity.constructor, entity[uniqueName], uniqueName);
                    if (entityWithUnique !== null && entityWithUnique[primary] !== entity[primary]) {
                        throw new Error(uniqueName + ' is not unique!');
                    }
                    if (entity[uniqueName] !== null) {
                        yield this.client.set(this.getUniqueKeyName(name, uniqueName) + ':' + entity[uniqueName], entity[primary]);
                    }
                }
            }
            const valuesToStore = [];
            for (const property of Object.keys(properties).map(key => properties[key])) {
                if (entity[property.name] !== null) {
                    let valueToStore = this.convertPropertyTypeToPrimitive(property, entity[property.name]);
                    if (hasOneRelations !== undefined && hasOneRelations[property.name]) {
                        const relatedEntity = this.metadata.getEntityMetadataFromName(hasOneRelations[property.name].entityType().name);
                        valueToStore = entity[property.name][relatedEntity.primary];
                        if (hasOneRelations[property.name].cascadeInsert && persistedEntity === null && entity[property.name] !== null) {
                            yield this.save(entity[property.name]);
                        }
                    }
                    valuesToStore.push(property.name);
                    valuesToStore.push(valueToStore);
                    if (property.searchable && entity[property.name]) {
                        yield this.client.sadd(this.getSearchableKeyName(name, property.name), this.getSearchableValuePrefix(entity[primary]) + entity[property.name].toLowerCase());
                    }
                    if (property.indexed) {
                        let value = entity[property.name];
                        if (hasOneRelations !== undefined && hasOneRelations[property.name] && entity[property.name] !== null) {
                            const relatedEntity = this.metadata.getEntityMetadataFromName(hasOneRelations[property.name].entityType().name);
                            value = entity[property.name][relatedEntity.primary];
                        }
                        if (value !== null) {
                            if (property.type === 'Date' || property.type === 'Number') {
                                yield this.client.zadd(this.getIndexNumberKeyName(name, property.name), this.convertPropertyTypeToPrimitive(property, entity[property.name]), entity[primary]);
                            }
                            else {
                                yield this.client.zadd(this.getIndexKeyName(name, property.name, value), '0', entity[primary]);
                            }
                        }
                    }
                }
            }
            yield this.client.hmset(hashKey, valuesToStore);
            if (canBeListed && persistedEntity === null) {
                yield this.client.rpush(this.getListKeyName(name), entity[primary]);
            }
            return null;
        });
    }
    count(entityType) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name } = this.metadata.getEntityMetadataFromType(entityType);
            const keyName = this.getListKeyName(name);
            return yield this.client.llen(keyName);
        });
    }
    list(entityType, where, limit, offset, orderBy) {
        return __awaiter(this, void 0, void 0, function* () {
            const ids = yield this.listIds(entityType, where, limit, offset, orderBy);
            const response = [];
            for (const id of ids) {
                response.push(yield this.getOne(entityType, id));
            }
            return response;
        });
    }
    search(entityType, condition, limit) {
        return __awaiter(this, void 0, void 0, function* () {
            const ids = yield this.searchIds(entityType, condition, limit);
            const response = [];
            const numberOfResult = (ids.length < limit) ? ids.length : limit;
            for (let index = 0; index < numberOfResult; index++) {
                response.push(yield this.getOne(entityType, ids[index]));
            }
            return response;
        });
    }
    searchIds(entityType, condition, limit) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name } = this.metadata.getEntityMetadataFromType(entityType);
            const key = this.getSearchableKeyName(name, condition.key);
            const value = this.getSearchableValuePrefix('*') + '*' + condition.value.toLowerCase() + '*';
            const response = [];
            let finishedScanning = false;
            let cursor = 0;
            while (!finishedScanning) {
                const scanResponse = (yield this.client.sscan(key, cursor, value));
                cursor = scanResponse.cursor;
                response.push(...scanResponse.data.map((id) => id.match(/.+?(?=\:_id_:)/g)[0]));
                if (cursor === 0 || response.length === limit) {
                    finishedScanning = true;
                }
            }
            return response;
        });
    }
    listIds(entityType, where, limit, offset, orderBy) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, canBeListed, properties } = this.metadata.getEntityMetadataFromType(entityType);
            if (!canBeListed) {
                throw new Error(entityType.name + ' can\'t be listed!');
            }
            const keyName = this.getListKeyName(name);
            let start = 0;
            let stop = -1;
            if (offset !== undefined) {
                start = offset;
            }
            if (limit !== undefined) {
                stop = start + limit - 1;
            }
            if (orderBy !== undefined && where === undefined) {
                const sortableKey = this.getIndexNumberKeyName(name, orderBy.field);
                if (orderBy.strategy === 'ASC') {
                    return yield this.client.zrange(sortableKey, start, stop);
                }
                else {
                    return yield this.client.zrevrange(sortableKey, start, stop);
                }
            }
            if (where !== undefined) {
                if (where.conditions.length === 0) {
                    throw new Error('Conditions can\'t be empty');
                }
                const scores = {};
                const equals = [];
                for (const condition of where.conditions) {
                    if (condition.comparator === '>') {
                        if (!scores[condition.key]) {
                            scores[condition.key] = { min: '-inf', max: '+inf' };
                        }
                        scores[condition.key].min = this.convertPropertyTypeToPrimitive(properties[condition.key], condition.value);
                    }
                    if (condition.comparator === '<') {
                        if (!scores[condition.key]) {
                            scores[condition.key] = { min: '-inf', max: '+inf' };
                        }
                        scores[condition.key].max = this.convertPropertyTypeToPrimitive(properties[condition.key], condition.value);
                    }
                    if (condition.comparator === '!=' || condition.comparator === '=') {
                        if (properties[condition.key] === undefined || !properties[condition.key].indexed) {
                            throw new Error('Property ' + condition.key + ' not found or not indexed');
                        }
                        equals.push(condition);
                    }
                }
                if (Object.keys(scores).length === 1 && equals.length === 0 && orderBy === undefined) {
                    const scoreKey = Object.keys(scores)[0];
                    return yield this.client.zrangebyscore(this.getIndexNumberKeyName(name, scoreKey), scores[scoreKey].min, scores[scoreKey].max, offset ? offset : 0, limit ? limit : -1);
                }
                if (equals.length === 1 && Object.keys(scores).length === 0 && orderBy === undefined && equals[0].comparator != "!=") {
                    const condition = equals[0];
                    return yield this.client.zrange(this.getIndexKeyName(name, condition.key, this.convertPropertyTypeToPrimitive(properties[condition.key], condition.value)), start, stop);
                }
                let luaOrderBy = undefined;
                if (orderBy !== undefined) {
                    for (const scoreKey in scores) {
                        if (scoreKey === orderBy.field) {
                            luaOrderBy = {
                                name: scoreKey,
                                min: String(scores[scoreKey].min),
                                max: String(scores[scoreKey].max),
                                strategy: orderBy.strategy,
                            };
                        }
                    }
                    if (luaOrderBy === undefined) {
                        luaOrderBy = {
                            name: orderBy.field,
                            strategy: orderBy.strategy,
                            min: "-inf",
                            max: "+inf",
                        };
                    }
                }
                let luaArgs = {
                    prefix: this.getIndexPrefix(name),
                    listKey: this.getListKeyName(name),
                    tempPrefix: 'temp:' + name + ':',
                    orderBy: luaOrderBy,
                    scores: Object.keys(scores).map((key) => ({
                        min: scores[key].min,
                        max: scores[key].max,
                        key,
                    })),
                    equals,
                    limit: limit ? limit : -1,
                    offset: offset ? offset : 0,
                    type: where.type,
                };
                return yield this.client.eval(fs.readFileSync(path.join(__dirname, './lua/complex.query.lua')), 0, JSON.stringify(luaArgs));
            }
            return yield this.client.lrange(keyName, start, stop);
        });
    }
    delete(entityType, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, uniques, canBeListed } = this.metadata.getEntityMetadataFromType(entityType);
            const hashKey = name + ':' + id;
            const persistedEntity = yield this.getOne(entityType, id);
            if (uniques) {
                yield this.dropUniqueKeys(persistedEntity);
            }
            if (canBeListed) {
                yield this.client.lrem(this.getListKeyName(name), 1, id);
            }
            yield this.dropIndexes(persistedEntity, id);
            yield this.dropSearchables(persistedEntity);
            yield this.client.del(hashKey);
        });
    }
    getOne(entityType, value, key) {
        return __awaiter(this, void 0, void 0, function* () {
            const entity = Object.create(entityType.prototype);
            const valueAsString = String(value);
            const { name, uniques, primary, properties, hasOneRelations } = this.metadata.getEntityMetadataFromType(entityType);
            let id;
            if (key !== undefined && key !== primary) {
                let indexKey;
                for (const uniqueName of uniques) {
                    if (uniqueName === key) {
                        indexKey = this.getUniqueKeyName(name, uniqueName);
                    }
                }
                if (indexKey === undefined) {
                    throw new Error(key + ' is not an unique field!');
                }
                id = yield this.client.get(indexKey + ':' + valueAsString);
            }
            else {
                id = valueAsString;
            }
            const hashKey = name + ':' + id;
            const result = yield this.client.hmget(hashKey, Object.keys(properties).map(key => properties[key].name));
            const propertiesArray = Object.keys(properties).map(key => properties[key]);
            let index = 0;
            for (const resultKey of result) {
                if (hasOneRelations !== undefined && hasOneRelations[propertiesArray[index].name] && resultKey !== null) {
                    entity[propertiesArray[index].name] = yield this.getOne(hasOneRelations[propertiesArray[index].name].entityType(), resultKey);
                }
                else {
                    entity[propertiesArray[index].name] = this.convertStringToPropertyType(propertiesArray[index], resultKey);
                }
                index++;
            }
            if (entity[primary] === null) {
                return null;
            }
            return entity;
        });
    }
    convertPropertyTypeToPrimitive(property, value) {
        if (property.type === 'Date') {
            return value.valueOf();
        }
        return String(value);
    }
    convertStringToPropertyType(property, value) {
        let convertedValue = value;
        switch (property.type) {
            case 'Boolean':
                convertedValue = value === 'true';
                break;
            case 'Number':
                convertedValue = Number(value);
                break;
            case 'Date':
                convertedValue = new Date(Number(value));
                break;
        }
        return convertedValue;
    }
    dropUniqueKeys(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, uniques } = this.metadata.getEntityMetadataFromInstance(entity);
            for (const uniqueName of uniques) {
                yield this.client.del(this.getUniqueKeyName(name, uniqueName) + ':' + entity[uniqueName]);
            }
        });
    }
    dropIndexes(entity, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const { properties } = this.metadata.getEntityMetadataFromInstance(entity);
            for (const property of Object.keys(properties).filter(key => properties[key].indexed).map(key => properties[key])) {
                yield this.dropIndex(entity, property, id);
            }
        });
    }
    dropIndex(entity, property, id) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, hasOneRelations } = this.metadata.getEntityMetadataFromInstance(entity);
            let value = entity[property.name];
            if (hasOneRelations !== undefined && hasOneRelations[property.name] && entity[property.name]) {
                const relatedEntity = this.metadata.getEntityMetadataFromName(hasOneRelations[property.name].entityType().name);
                value = entity[property.name][relatedEntity.primary];
            }
            if (property.type === 'Date' || property.type === 'Number') {
                yield this.client.zrem(this.getIndexNumberKeyName(name, property.name), id);
            }
            else {
                yield this.client.zrem(this.getIndexKeyName(name, property.name, value), id);
            }
        });
    }
    dropSearchables(entity) {
        return __awaiter(this, void 0, void 0, function* () {
            const { name, properties, primary } = this.metadata.getEntityMetadataFromInstance(entity);
            for (const property of Object.keys(properties).map(key => properties[key])) {
                if (property.searchable && entity[property.name]) {
                    yield this.client.srem(this.getSearchableKeyName(name, property.name), this.getSearchableValuePrefix(entity[primary]) + entity[property.name].toLowerCase());
                }
            }
        });
    }
    getIndexNumberKeyName(entityName, indexName) {
        return this.getIndexPrefix(entityName) + indexName;
    }
    getIndexKeyName(entityName, indexName, indexValue) {
        return this.getIndexPrefix(entityName) + indexName + ':' + indexValue;
    }
    getIndexPrefix(entityName) {
        return entityName + ':index:';
    }
    getListKeyName(entityName) {
        return entityName + ':list';
    }
    getUniqueKeyName(entityName, uniqueName) {
        return entityName + ':unique:' + uniqueName;
    }
    getSearchableKeyName(entityName, fieldName) {
        return entityName + ':search:' + fieldName;
    }
    getSearchableValuePrefix(id) {
        return id + ':_id_:';
    }
}
exports.Redisk = Redisk;
